#include "Wolf.h"
#include "Definitions.h"

Wolf::Wolf(World *w){
   
  my_world = w;

  SetParameters();
  pos[2] = my_world->GetHeight(pos[0], pos[1]);
}


Wolf::Wolf(float x, float y, World *w){
   
  my_world = w;

  SetParameters();
  
  pos[0] = x;
  pos[1] = y;
  pos[2] = my_world->GetHeight(pos[0], pos[1]);
   

}


void Wolf::SetParameters(){

  int zufall;
   
  pos[0] = PaeRand(0.0,99.0);
  pos[1] = PaeRand(0.0,99.0);
   
  angle = PaeRand(0.0,360.0);
   
  dir[0] = sin(angle/360.0*M_PI);
  dir[1] = cos(angle/360.0*M_PI);
  dir[2] = 0.0;
   
  (void) PaeRenorm3(dir);
   
  up[0] = 0.0;
  up[1] = 0.0;
  up[2] = 1.0;
   
  acc[0] = 0.0;
  acc[1] = 0.0;
  acc[2] = 0.0;
   
  col[0] = 0.3;
  col[1] = 0.3;
  col[2] = 0.3;
   
  jagd_instinkt = PaeRand(0.0,0.1);

  hunger = PaeRand(0.0,1.0);
   
  maxausdauer = (int) PaeRand(25.0, 50.0);
  ausdauer = maxausdauer;
   
  nextwolf = NULL;
   
  zufall = (int) PaeRand(0.0,2.9);
  switch (zufall) {
  case 0:
	state = SCHLAF;
	schlafzeit = SCHLAFZEIT + (int) PaeRand(-1.0*SCHLAFDELTA, 1.0*SCHLAFDELTA);
	break;
  case 1: 
	state = WANDERN;
	break;
  case 2: 
	state = JAGD;
	hunger = 0.95;
	break;
  default:
	cout << "Fehler in Wolf::SetParameters: falscher status!" << endl;
  }
}


void Wolf::SetPosition(GLfloat x, GLfloat y, GLfloat z){
  pos[0] = x;     
  pos[1] = y;
  pos[2] = z;
}


void Wolf::SetDirection(GLfloat x, GLfloat y, GLfloat z){
  dir[0] = x;
  dir[1] = y;
  dir[2] = z;
}

void Wolf::SetColor(GLfloat r, GLfloat g, GLfloat b){
  col[0] = r;
  col[1] = g;
  col[2] = b;
}





void Wolf::SetVisColor(GLfloat r, GLfloat g, GLfloat b){
  vis_col[0] = r;
  vis_col[1] = g;
  vis_col[2] = b;
}




void Wolf::Think(void){
  float pos_sheep[3];
  float len;
  float pos_next[3];
  PaeSheep *nearest_sheep;

  PaeSheep **prey;
  float probab[5];
  float sum,zufall;
  int preycount,i;
  /*
   * Zunaechst muss mal festgelegt werden, in welchem Status
   * (Variable **state**) sich der Wolf befindet
   *
   * Zentrale Variablen sind *hunger* und *ausdauer*. Wenn
   * die Ausdauer fehlt, kann der Wolf nichts anderes machen, 
   * als schlafen. Ansonsten entscheidet sein Hunger, ob er 
   * WANDERN oder JAGD als Zustand annimmt.
   */


  switch(state){
  case JAGD:
	if(ausdauer <= 0){
	  ausdauer = 0;
	  state = SCHLAF;
	  schlafzeit = SCHLAFZEIT+(int)PaeRand(-SCHLAFDELTA, SCHLAFDELTA);
	}
	break;
  case SCHLAF:
    if(schlafzeit <= 0){
	  if(hunger > 0.9){
		state = JAGD;
	  } else {
		state = WANDERN;
	  }
	}
	break;
  case WANDERN:
    if(hunger > 0.9){
      state = JAGD;
    }
	break;
  case FRESSEN:
	break;
  default:
	cout << "falscher State in Wolf::Move" << endl;
	break;
  }


  /*
   * Jetzt wird das Denken (je nach State) eingeschaltet
   */
  switch(state){
  case JAGD:
    this->SetVisColor(1.0,0.0,0.0);//ein jagender wolf ist rot
    
	/*
	 * suche die fuenf naehesten schafe
	 */
    prey = my_world->GetFiveNearestSheep(this);

	// keine Schafe mehr? Umherirren!!
	if(prey[0] == NULL){
	  acc[0] = PaeRand(-0.2,0.2);
	  acc[1] = PaeRand(-0.2,0.2);
	  acc[2] = 0.0;
	  return;
	}

	// Schafe zur Auswahl bereit machen
	sum = 0.0;
	preycount = 0;
	for(i=0;i<5;i++){
	  if(prey[i] == NULL){
		break;
	  }
	  probab[i] = 1.0/(Distance(this,prey[i])+0.1);
	  sum+=probab[i];
	  preycount++;
	}
	/*
	 * jetzt haben wir in probab alle Auswahlprobs., in sum die
	 * summe und in preycount die Anzahl der Schafe in der
	 * Umgebung
	 * nun wird eines zum Jagen ausgesucht
	 */
	zufall = PaeRand(0.0,1.0);
	nearest_sheep = prey[preycount-1];
	for(i=0;i<preycount;i++){
	  if(zufall > probab[i]/sum){
		zufall -= probab[i]/sum;
	  } else {
		nearest_sheep = prey[i];
		break;
	  }
	}

	nearest_sheep->GetPosition(pos_sheep);

	pos_next[0] = pos[0]+dir[0]*v;
	pos_next[1] = pos[1]+dir[1]*v;
	pos_next[2] = pos[2]+dir[2]*v;

    acc[0] = (pos_sheep[0]-pos_next[0])*jagd_instinkt;
    acc[1] = (pos_sheep[1]-pos_next[1])*jagd_instinkt;
    acc[2] = (pos_sheep[2]-pos_next[2])*jagd_instinkt;
    
    len = PaeRenorm3(acc);
    
    //ist das tier sehr nahe am ziel, bremst es ab
    if(len < 1.0){
      v = MAXSPEED/2.0;
    }
	break;
  case SCHLAF:
    this->SetVisColor(0.0,0.0,1.0);//ein schlafender wolf ist blau
        
    acc[0] = 0.0;
    acc[1] = 0.0;
    acc[2] = 0.0;

    v = 0.0;
	break;
  case WANDERN:
    this->SetVisColor(1.0,0.0,1.0);//ein wandernder Wolf ist lila
    acc[0] = PaeRand(-0.2,0.2);
    acc[1] = PaeRand(-0.2,0.2);
    acc[2] = 0.0;
        
	v = 0.2;
	break;
  case FRESSEN:
    this->SetVisColor(0.9,0.8,0.0);//ein fressender Wolf ist orange
        
	break;
  default:
	cout << "falscher State in Wolf::Move" << endl;
	break;
  }

}


void Wolf::Move(void){
  float len, dist;
  float pos_sheep[3];
  float dx,dy,dz;
  float senkrecht[3];
  
  PaeSheep *nearest_sheep;

  switch (state) {
  case SCHLAF:
  case FRESSEN:
	v = 0.0;
	// sicher ist sicher: Position anpassen
	pos[2] = my_world->GetHeight(pos[0], pos[1]);
	break;

  case JAGD:
  case WANDERN:
	UtilCrossProduct(dir,acc,senkrecht);
	UtilCrossProduct(senkrecht,dir,acc);
    //Aenderung der Richtung
    dir[0] += acc[0];
    dir[1] += acc[1];
    dir[2] += acc[2];
    
    (void)PaeRenorm3(dir);
	angle = asin(-dir[1]);
	angle = angle*180.0/3.1415926;
	if(dir[0] > 0.0){
	  if (angle > 0.0){
		angle = 180-angle;
	  } else { 
		angle = -angle-180.0;
	  }
	}

	/* Jagende Woelfe werden schneller */
	if (state == JAGD){
	  v*=1.3;
	  UtilClamp(&v,MINSPEED,MAXSPEED);
	}

	/* 
	 * neue x-y-Position berechnen und z-Position anpassen
	 */

    pos[0] += dir[0] * v;
    pos[1] += dir[1] * v;
    pos[2] = my_world->GetHeight(pos[0], pos[1]);
    
	break;
  default:
	cout << "falscher State in Wolf::Move" << endl;
    break;
  }
}



void Wolf::Act(void){
  float len, dist;
  float pos_sheep[3];
  float location[3];
  float dx,dy,dz;  
  PaeSheep *nearest_sheep;
  
  switch (state) {
  case SCHLAF:
    // Schlafuhr runterzaehlen
    schlafzeit--;
   
    // Schlafen bringt schnell Ausdauer
	ausdauer += 2;
	UtilClamp(&ausdauer,0,maxausdauer);

    // Man wird etwas hungriger
    hunger += PaeRand(0.0,0.01);
	UtilClamp(&hunger,0.0,1.0);
	break;

  case WANDERN:
    // Wandern bringt etwas Ausdauer
	ausdauer += 1;
	UtilClamp(&ausdauer,0,maxausdauer);

    // Man wird etwas hungriger
    hunger += PaeRand(0.0,0.02);
	UtilClamp(&hunger,0.0,1.0);
	break;

  case JAGD:
	ausdauer-=2;

	/* Ist die Jagd erfolgreich? */
    nearest_sheep = my_world->GetNearestSheep(this);
	if (nearest_sheep == NULL){
	  return;
	}
	nearest_sheep->GetPosition(pos_sheep);

	dx = pos_sheep[0] - pos[0];
	dy = pos_sheep[1] - pos[1];
	dz = pos_sheep[2] - pos[2];
	dist =dx*dx+dy*dy+dz*dz;

	// Wenn der Wolf bei einem Schaf ist, versucht er, es zu reissen
    if(dist < JAGDABSTAND){
      if(my_world->KillSheep(pos_sheep)){
        // hat es erfolgreich gejagt, frisst er
        state = FRESSEN;
      }
    }
	break;
  case FRESSEN:
	/*
	 * frisst an der Position, wo er sich gerade
     * aufhaelt. Wenn es dort nichts mehr gibt, wandert
	 * er weiter oder schlaeft.
	 */ 
	this->GetPosition(location);
	if(my_world->EatFlesh(location,0.1)){
	  hunger *=0.9;
	  break;
	} 
	  
	if (ausdauer > maxausdauer/2.0){
		state = WANDERN;
	} else {
	  state = SCHLAF;
	  schlafzeit = SCHLAFZEIT+(int)PaeRand(-SCHLAFDELTA, SCHLAFDELTA);
	}

	break;
  default:
	cout << "falscher State in Wolf::Act" << endl;
  }
}



void Wolf::Draw(int visual){
  glPushMatrix();
    glTranslatef(pos[0], pos[1], pos[2]+0.5);
	glRotatef(angle,0.0,0.0,1.0);
	if (visual == 0){
	  glColor3fv(col);
	} else {
	  glColor3fv(vis_col);
	}
	glScalef(1.5,0.5,0.5);
    glutSolidSphere(1.0,10,10);
  glPopMatrix();
}

void Wolf::GetPosition(float *p){

  p[0] = pos[0];
  p[1] = pos[1];
  p[2] = pos[2];
}
  
void Wolf::GetDirection(float *d){

  d[0] = dir[0];
  d[1] = dir[1];
  d[2] = dir[2];
}
